/*
 * -------------------------------------------------------------------
 *
 * Copyright (c) 2010, Cardinal Peak, LLC.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1) Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2) Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 * 
 * 3) Neither the name of Cardinal Peak nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
 * CARDINAL PEAK, LLC BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * 
 * -------------------------------------------------------------------
 * 
 * Simulates a delta sigma converter.  Outputs the simulated sample
 * values to a file.  NOTE: values output are not "AD" bitstream 0's
 * and 1's.  Outputs are the values that can be derived from the
 * pulses output by the delta sigma modulator.  In other words, the
 * output of the simulation is the output of the filter and decimation
 * stage of the converter.
 */

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <inttypes.h>

#define INTEGRAL_FACTOR   20    // no need to change this for cosine inputs
#define OVERSAMPLE_FACTOR 64
#define MAX_SIGNAL_BW     4000.0

void fir_filt_init(char *filename);
double fir_filt( double d, int compute_sample );

int main(int argc, char *argv[] )
{
	FILE *out_fd;
	double t, delta_t, y, v, time_stop, lnm1, ln, result;
	int i, bit_cnt, compute_sample;

	// Simulation time starts at 0 and ends at time_stop.  taps.txt
	// contains the FIR filter taps and out.txt contains the sample
	// values in a two column text file
	if( argc != 4 ){
		fprintf(stderr, "usage is: deltasigma time_stop taps.txt out.txt\n");
		exit(1);
	}

	time_stop = atof( argv[1] );
	fir_filt_init( argv[2] );
	out_fd = fopen( argv[3], "w");

	// Initialize the simulation time increment
	delta_t = 1.0 / (MAX_SIGNAL_BW * 2.0 * OVERSAMPLE_FACTOR * INTEGRAL_FACTOR);

	// Initialize some other variables
	t = 0;       // time
	y = 0;       // integrator's value
	ln = 1;      // l(n)
	bit_cnt = 0; // a loop counter

	while(t < time_stop){
		lnm1 = ln; // update l(n-1) to l(n)
		// This loop models the summer and the integrator.  INTEGRAL_FACTOR
		// points are used to approximate the integral value accumulated
		// between output pulses
		for(i = 0; i < INTEGRAL_FACTOR; i++){
			// v is the input signal
		  v = 0.5 * cos(2*M_PI*3765.31*t) + 0.5*cos(2*M_PI*730.18*t);
			y += (v - lnm1) * delta_t; // integration
			t += delta_t;
		}

		// This models the quantizer
		if( y > 0 ) ln = 1;
		else ln = -1;
		bit_cnt++;

		// Output the result.  Because we oversampled by OVERSAMPLE_FACTOR,
		// we can sub sample the result.  The compute_sample variable controls
		// whether a value is computed and hence output to the file
		
		if( bit_cnt % OVERSAMPLE_FACTOR == 0 ) compute_sample = 1;
		else compute_sample = 0;
		result = fir_filt(ln, compute_sample);

		//Uncomment the line below to output the modulator pulse stream
		//instead of decimated samples from the lowpass filtered signal
		//result = ln; compute_sample = 1;

		if( compute_sample && (t < time_stop) ){
		  fprintf(out_fd, "%13.10e  %13.10e\n", t, result);
		}	
	}

	fclose( out_fd );
}

/* The two routines below implement the FIR filter */

double *m_taps;
double *m_data_r;
int m_num_taps;

void fir_filt_init(char *filename)
{
	int i;
	FILE *fd_in;

	fd_in = fopen(filename, "r");
	if( fd_in == NULL ){
		fprintf(stderr, "ERROR: could not open tap file\n");
		exit(1);
	}

	fscanf(fd_in, "%d", &m_num_taps);

	m_taps = (double *) malloc(m_num_taps * sizeof(double));
	m_data_r = (double *) malloc(m_num_taps * sizeof(double));

	for(i = 0; i < m_num_taps; i++){
		m_data_r[i] = 0;
		fscanf(fd_in, "%lf", &m_taps[i]);
	}

	fclose(fd_in);
}

double fir_filt( double d, int compute_sample ){
	int i;
	double d1, result;

	for(i = m_num_taps - 1; i >= 1; i--){
		m_data_r[i] = m_data_r[i-1];
	}	
	m_data_r[0] = d;

	if( !compute_sample ) return( -1e50 );

	result = 0;
	for(i = 0; i < m_num_taps; i++){
		d1 = m_data_r[i] * m_taps[i];
		result += d1;
	}

	return result;
}
